/* upload.c
 * Stefano Salvi - 8/12/2002
 * Programma per scaricare sulla scheda GPC2 un file .hex da riga di comando.
 * per default scarica il file "protocol.hex" ed usa la seriale "/dev/ttyS0"
 * Accetta i seguenti paramentri:
 *  -f<file sa scaricare> indica che file '.hex' scaricare (default protocol.hex)
 *  -s<seriale> indica il percorso per il dispositivo seriale (default /dev/ttyS0)
 *  -d non far partire il file scaricato (scarica e basta)
 *  -r non scaricare alcun file (avvia il programma sulla scheda)
 */

#include <sched.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

int com;	// handle della seriale

/* Invia una stringa sulla seriale, verso la scheda
 */
void schPuts (const char *s)
{
  write (com, s, strlen (s));
}

/* Invia un singolo carattere sulla seriale, alla scheda
 */
void schPutc (const char c)
{
  write (com, &c, 1);
}

/* Ritorna un carattere dalla pipe della scheda
 * con un timeout di 3 secondi
 * Se il carattere non arriva, ritorna -1;
 */
#define SCHWAIT	5
int schGetc ()
{
fd_set set;
struct timeval timeout;
unsigned char c;

  /* Initialize the file descriptor set. */
  FD_ZERO (&set);
  FD_SET (com, &set);
     
  /* Initialize the timeout data structure. */
  timeout.tv_sec = SCHWAIT;
  timeout.tv_usec = 0;
  /* `select' returns 0 if timeout, 1 if input available, -1 if error. */
  if (!select (FD_SETSIZE, &set, NULL, NULL, &timeout))
  {
    schPutc (0x1b);	// Se si e' interrotto per errore, interrompe il comando
    printf ("La scheda non risponde\n");
    exit (-1);
  }
  read (com,&c,1);
  return c;
}

/* Ritorna una linea letta dalla scheda (terminata da \r),
 * lunga al massimo 'max' caratteri.
 * Non inserisce il '\r' o '\n' nella stringa
 * Ritorna:
 *   true -> errore;
 *   false -> OK;
 */
int schGets (char *str,int max)
{
int rec;
  do {
    rec = schGetc ();
    if (rec != '\n' && rec != '\r')
      *str++ = rec;
  } while (rec >= 0 && rec != '\r' && --max);
  *str = 0;
  return rec < 0;
}

/* Attende un particolare carattere ('c') dalla scheda.
 * Se arriva il carattere 'c' ritorna 'false'.
 * Se avviene un timeout (nessun carattere per almeno 5 sec) ritorna 'true'
 */
int schWaitc (char c)
{
int rec;
  do {
    rec = schGetc ();
  } while (rec >= 0 && rec != c);
  return rec < 0;
}

/* Invia alla scheda un comando 'go' all'indirizzo 'where'
 */
void doRun (int where)
{
char gobuf [6];
int i;

    sprintf (gobuf,"G%04X",where);
    schPuts (gobuf);	// Comando 'go'
    // Risucchia l'eco dei cinque caratteri stampati
    for (i = 0; i<5; i++)
      if (schGetc () < 0)
        break;
}

/* Invia alla scheda un comando 'go' all'indirizzo 0x8050 (inizio programmi)
 */
void runStart ()
{
  doRun (0x8050);
}

/* Attende il prompt della scheda ('*').
 * Ritorna -1 se il prompt non arriva (timeout)
 * Ritorna 0 se il prompt arriva
 */
int schPrompt (void)
{ 
  if (schWaitc ('*')) {
    return -1;
  } 
  return 0; 
}

#define BUFLEN	512

/* Attende la fine del comando di upload (o di una sua parte)
 */
void waitEndOfCommand ()
{
char buf [BUFLEN];	// Buffer per leggere il file
  if (schWaitc ('\r') ||	// scarta il comando
    schGets (buf,79) ||
    schPrompt ())
  {
    fprintf (stderr, "La scheda non risponde\n");
    exit (3);
  } else {
    if (buf [0] == '?') {
      fprintf (stderr, "Errore nel file '%s'\n" , buf);
      exit (3);
    }
  }
}

/* Invia il file il cui nome e' '*file' (che deve essere in formato 'hex intel')
 * alla scheda
 */
void doUpload (char *file)
{
int hexfile;		// File handle del file da inviare
char buf [BUFLEN];	// Buffer per leggere il file
int letti;		// Dimensione dell'ultimo blocco letto
int i;			// indice nel buffer
int l;			// numero di linee
int f;			// Indice della stringa trovata
char *endHex = ":00000001FF\r\n";
    
  hexfile = open (file,O_RDONLY);	// Apre il file
  if (hexfile == EOF)
  {
    fprintf (stderr, "Non trovo il file '%s' da scaricare\n", file);
    exit (3);
  }

  schPuts ("L\r");	// Comando 'upload'
  l = 0;
  f = 0;

  /* Ciclo di lettura del file e di invio alla scheda
   */
  do {
    letti = read (hexfile,buf,BUFLEN);	// Legge il prossimo blocco
    for (i = 0;i < letti;i++)		// Ciclo per i byte letti
    {
      if (buf [i] != '\r')		// Elimina gli eventuali \r (CR)
      {
	if (buf [i] == '\n') {		// Se incontra un \n (LF)
	  write (com,"\r",1);		// Ci mette prima un \r (CR)
//	  schPuts ("\r");
        }
	write (com,buf + i,1);		// Invia il carattere dal file
        if (endHex [f] != '\r')		// Se non ha ancora trovato al stringa
        {
          if (endHex [f] == buf [i])	// Se la stringa corrisponde ancora
          {
            f ++;			// Va al prossimo carattere
          } else {
            f = 0;			// Se no, torna a 0
          }
        }
	if (buf [i] == '\n') {		// Se ha spedito un \n
          if (l ++ > 30)		// Ogni 30 righe 'chiude e riapre'
          {
            l = 0;
            if (endHex [f] != '\r')     // Se non ha ancora inviato 'fine file hex'
            {
	      write (com,endHex,strlen (endHex)); // Invia il carattere dal file
              waitEndOfCommand ();	// attendo conferma
              schPuts ("L\r");		// Nuovo comando 'upload'

              putchar ('+');
              fflush (stdout);
            } else {
              putchar ('*');
              fflush (stdout);
            }
          } else {
            if (endHex [f] == '\r')     // Se non ha ancora inviato 'fine file hex'
            {
              putchar ('*');
              fflush (stdout);
            } else {
              putchar ('.');
              fflush (stdout);
            }
          }
          usleep (10000);		// fa una pausa (la scheda e' lenta....)
          f = 0;
        }
      }
    }
  } while (letti==BUFLEN);		// Fintantoche' ha letto dati

  printf ("\nFine upload\n");
  fflush (stdout);

  waitEndOfCommand ();
}

/* Apre la seriale indicata dal parametro e la imposta in modo non canonico
 */
int setupCom (char *comname)
{
struct termios tattr;

    com = open (comname,O_RDWR | O_SYNC);
    if (com == -1)
    {
	perror ("Non posso aprire la seriale");
        exit (2);
    }

    /* Set the funny terminal modes. */
    tcgetattr (com, &tattr);
    tattr.c_iflag &= ~(INLCR|IGNCR|ICRNL|IXON); /*  */
    tattr.c_oflag &= ~(OPOST|ONLCR|OCRNL|ONLRET); /*  */
    tattr.c_cflag = CS8 | CREAD | CLOCAL; 
    tattr.c_lflag &= ~(ICANON|ECHO); /* Clear ICANON and ECHO. */
    tattr.c_cc[VMIN] = 1;
    tattr.c_cc[VTIME] = 0;
    cfsetispeed (&tattr, B9600);
    cfsetospeed (&tattr, B9600);
    tcsetattr (com, TCSAFLUSH, &tattr);

    return com;
}

char fileToLoad [] = "protocol.hex";
char ttyName [] = "/dev/ttyS0";

/* Stampa messaggio di errore per i parametri e termina
 */
void parerr (char *nome)
{
  fprintf (stderr, "Errore. Uso:\n%s [-f<file sa scaricare>] [-s<seriale>] [-d] [-r]\n"
    "-f<file sa scaricare> indica che file '.hex' scaricare (default %s)\n"
    "-s<seriale> indica il percorso per il dispositivo seriale (default %s)\n"
    "-d non far partire il file scaricato (scarica e basta)\n"
    "-r non scaricare alcun file (avvia il programma sulla scheda)\n",
    nome, fileToLoad, ttyName);
  exit (1);
}

/* Programma principale */
int main (int argc, char **argv)
{
int i;
int download = -1;		// Se 'true', esegue il download
int start = -1;			// Se 'true', avvia il programma
char *toLoad = fileToLoad;	// Nome del file da scaricare
char *ttyN = ttyName;		// Nome del dispositivo della seriale

  // Loop di analisi dei parametri
  for (i = 1; i < argc; i++) {
    if (argv [i][0] == '-') {
      switch (argv [i][1]) {
        case 'f' :
          toLoad = argv [i] + 2;
          break;
        case 's' :
          ttyN = argv [i] + 2;
          break;
        case 'd' :
          start = 0;
          break;
        case 'r' :
          download = 0;
          start = -1;
          break;
        default :
          parerr (argv [0]);
      }
    } else {
      parerr (argv [0]);
    }
  }

  com = setupCom (ttyN);		// Imposta la seriale

  if (download) {
    doUpload (toLoad);			// Se richieso, scarica il file nella scheda
  }

  if (start) {
    runStart ();			// Se richiesto, fa paritre il programma sulla scheda
  }

  return 0;
}